#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "block.h"
#include "area.h"
#include "file_object.h"
#include "file_node.h"

typedef struct _FileName *FileName;

struct file_node {
  unsigned int flags;
  int user_id;
  int group_id;
  int access_time;   /* ǸΥե륢(ɤߤor¹) */
  int modified_time; /* ǸΥեѹ(񤭹) */
  int changed_time;  /* Ǹinodeѹ(ե°ʤɤѹ) */

  FileName name;
  FileObject object;
};

static FileName FileNode_GetName(FileNode node);
static FileName FileNode_SetName(FileNode node, FileName name);
static FileName FileNode_NameCreate(FileNode node, BlockPool pool);
static FileName FileNode_NameDestroy(FileNode node);

/******** for FileNode operation ********/

BlockPool FileNode_GetPool(FileNode node)
{
  return Area_GetPool((Area)node);
}
static int FileNode_Read(FileNode node, struct file_node *info)
{
  return Area_Read((Area)node, 0, sizeof(*info), info);
}

static int FileNode_Write(FileNode node, struct file_node *info)
{
  FileNode_SetChangedTimeNow(node);
  return Area_Write((Area)node, 0, sizeof(*info), info);
}

unsigned int FileNode_GetFlags(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.flags;
}

unsigned int FileNode_SetFlags(FileNode node, unsigned int flags)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.flags = flags;
  FileNode_Write(node, &info);
  return info.flags;
}

int FileNode_GetUserID(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.user_id;
}

int FileNode_SetUserID(FileNode node, int id)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.user_id = id;
  FileNode_Write(node, &info);
  return info.user_id;
}

int FileNode_GetGroupID(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.group_id;
}

int FileNode_SetGroupID(FileNode node, int id)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.group_id = id;
  FileNode_Write(node, &info);
  return info.group_id;
}

int FileNode_GetAccessTime(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.access_time;
}

int FileNode_SetAccessTime(FileNode node, int t)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.access_time = t;
  FileNode_Write(node, &info);
  return info.access_time;
}

int FileNode_SetAccessTimeNow(FileNode node)
{
  return FileNode_SetAccessTime(node, time(NULL));
}

int FileNode_GetModifiedTime(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.modified_time;
}

int FileNode_SetModifiedTime(FileNode node, int t)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.modified_time = t;
  FileNode_Write(node, &info);
  return info.modified_time;
}

int FileNode_SetModifiedTimeNow(FileNode node)
{
  return FileNode_SetModifiedTime(node, time(NULL));
}

int FileNode_GetChangedTime(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.changed_time;
}

int FileNode_SetChangedTime(FileNode node, int t)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.changed_time = t;
#if 0
  FileNode_Write(node, &info);
#else /* ̵¥롼פɻ */
  Area_Write((Area)node, 0, sizeof(info), &info);
#endif
  return info.changed_time;
}

int FileNode_SetChangedTimeNow(FileNode node)
{
  return FileNode_SetChangedTime(node, time(NULL));
}

static FileName FileNode_GetName(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.name;
}

static FileName FileNode_SetName(FileNode node, FileName name)
{
  struct file_node info;
  FileNode_Read(node, &info);
  info.name = name;
  FileNode_Write(node, &info);
  return info.name;
}

FileObject FileNode_GetObject(FileNode node)
{
  struct file_node info;
  FileNode_Read(node, &info);
  return info.object;
}

FileObject FileNode_SetObject(FileNode node, FileObject object)
{
  struct file_node info;
  FileObject old;

  FileNode_Read(node, &info);
  old = info.object;
  if (old)
    FileObject_DelReference(old);
  info.object = object;
  if (object)
    FileObject_AddReference(object);
  FileNode_Write(node, &info);

  if (old && (FileObject_GetReference(old) == 0)) {
    FileObject_Destroy(old);
  }

  return info.object;
}

FileObject FileNode_ResetObject(FileNode node)
{
  return FileNode_SetObject(node, NULL);
}

FileNode FileNode_Create(BlockPool pool, const char *name)
{
  FileNode node;
  struct file_node info;
  time_t t;

  node = (FileNode)Area_Create(pool);
  Area_Resize((Area)node, sizeof(info));

  memset(&info, 0, sizeof(info));

  t = time(NULL);
  info.user_id = 1001;
  info.group_id = 1001;
  info.access_time = t;
  info.modified_time = t;
  info.changed_time = t;

  info.name   = NULL;
  info.object = NULL;

  FileNode_Write(node, &info);

  FileNode_NameCreate(node, pool);
  FileNode_NameWrite(node, name);

  return node;
}

FileNode FileNode_Destroy(FileNode node)
{
  FileNode_NameDestroy(node);
  FileNode_ResetObject(node);
  Area_Destroy((Area)node);
  return NULL;
}

/******** for FileName operation ********/

int FileNode_NameGetSize(FileNode node)
{
  FileName name = FileNode_GetName(node);
  return Area_GetSize((Area)name);
}

int FileNode_NameRead(FileNode node, int size, char *buffer)
{
  FileName name = FileNode_GetName(node);
  return Area_Read((Area)name, 0, size, buffer);
}

int FileNode_NameWrite(FileNode node, const char *buffer)
{
  FileName name = FileNode_GetName(node);
  int size;
  size = strlen(buffer) + 1;
  Area_Resize((Area)name, size);
  return Area_Write((Area)name, 0, size, buffer);
}

int FileNode_NameCompare(FileNode node, const char *buffer)
{
  FileName name = FileNode_GetName(node);
  int size;
  size = strlen(buffer) + 1;
  return Area_Compare((Area)name, 0, size, buffer);
}

static FileName FileNode_NameCreate(FileNode node, BlockPool pool)
{
  return FileNode_SetName(node, (FileName)Area_Create(pool));
}

static FileName FileNode_NameDestroy(FileNode node)
{
  return FileNode_SetName(node, (FileName)Area_Destroy((Area)FileNode_GetName(node)));
}
